"""
Pytest Hot Reloading plugin
"""
from __future__ import annotations

import os
import sys
from typing import TYPE_CHECKING, Callable

from pytest_hot_reloading.client import PytestClient

# this is modified by the daemon so that the pytest_collection hooks does not run
i_am_server = False

seen_paths = set()

if TYPE_CHECKING:
    from pytest import Config, Item, Session


def pytest_addoption(parser) -> None:
    group = parser.getgroup("daemon")
    group.addoption(
        "--daemon",
        action="store_true",
        default=False,
        help="Start the daemon",
    )
    group.addoption(
        "--daemon-port",
        action="store",
        default=4852,
        help="The port to use for the daemon. You generally shouldn't need to set this.",
    )
    group.addoption(
        "--pytest-name",
        action="store",
        default="pytest",
        help="The name of the pytest executable or module",
    )
    group.addoption(
        "--daemon-timeout",
        action="store",
        default=(5 * 60),
        help="The timeout in seconds to wait on a test suite to finish",
    )
    group.addoption(
        "--daemon-watch-globs",
        action="store",
        default="./*.py",
        help="The globs to watch for changes. This is a colon separated list of globs.",
    )
    group.addoption(
        "--daemon-ignore-watch-globs",
        action="store",
        default="./.venv/*",
        help="The globs to ignore for changes. This is a colon separated list of globs.",
    )


# list of pytest hooks
# https://docs.pytest.org/en/stable/reference.html#_pytest.hookspec.pytest_addhooks


def pytest_cmdline_main(config: Config) -> None:
    """
    This hook is called by pytest and is one of the first hooks.
    """
    # early escapes
    if config.option.collectonly:
        return
    if i_am_server:
        return
    _plugin_logic(config)
    # dont do any more work. Don't let pytest continue
    return 0  # status code 0


def monkey_patch_jurigged_function_definition():
    import jurigged.codetools as jurigged_codetools

    OrigFunctionDefinition = jurigged_codetools.FunctionDefinition

    import ast

    class NewFunctionDefinition(OrigFunctionDefinition):
        def reevaluate(self, new_node, glb):
            new_node = self.apply_assertion_rewrite(new_node, glb)
            return super().reevaluate(new_node, glb)

        def apply_assertion_rewrite(self, ast_func, glb):
            from _pytest.assertion.rewrite import AssertionRewriter

            nodes: list[ast.AST] = [ast_func]
            while nodes:
                node = nodes.pop()
                for name, field in ast.iter_fields(node):
                    if isinstance(field, list):
                        new: list[ast.AST] = []
                        for i, child in enumerate(field):
                            if isinstance(child, ast.Assert):
                                # Transform assert.
                                new.extend(
                                    AssertionRewriter(glb["__file__"], None, None).visit(child)
                                )
                            else:
                                new.append(child)
                                if isinstance(child, ast.AST):
                                    nodes.append(child)
                        setattr(node, name, new)
                    elif (
                        isinstance(field, ast.AST)
                        # Don't recurse into expressions as they can't contain
                        # asserts.
                        and not isinstance(field, ast.expr)
                    ):
                        nodes.append(field)
            return ast_func

    # monkey patch in new definition
    jurigged_codetools.FunctionDefinition = NewFunctionDefinition


def setup_jurigged(config: Config):
    def _jurigged_logger(x: str) -> None:
        """
        Jurigged behavior is to both print and log.

        By default this creates duplicated output.

        Pass in a no-op logger to prevent this.
        """

    import jurigged

    pattern = _get_pattern_filters(config)
    # TODO: intelligently use poll versus watchman (https://github.com/JamesHutchison/pytest-hot-reloading/issues/16)
    jurigged.watch(pattern=pattern, logger=_jurigged_logger, poll=True)


def _plugin_logic(config: Config) -> None:
    """
    The core plugin logic. This is where it splits based on whether we are the server or client.

    In either case, the pytest logic will not continue after this.
    """
    # if daemon is passed, then we are the daemon / server
    # if daemon is not passed, then we are the client
    daemon_port = int(config.option.daemon_port)
    if config.option.daemon:
        # pytest prints out "collecting ...". The leading \r prevents that
        print("\rStarting daemon...")
        setup_jurigged(config)

        from pytest_hot_reloading.daemon import PytestDaemon

        daemon = PytestDaemon(daemon_port=daemon_port)

        daemon.run_forever()
    else:
        pytest_name = config.option.pytest_name
        client = PytestClient(daemon_port=daemon_port, pytest_name=pytest_name)
        # find the index of the first value that is not None
        for idx, val in enumerate(
            [
                x.endswith(pytest_name) or x.endswith(f"{pytest_name}/__main__.py")
                for x in sys.argv
            ]
        ):
            if val:
                pytest_name_index = idx
                break
        else:
            print(sys.argv)
            raise Exception(
                "Could not find pytest name in args. "
                "Check the configured name versus the actual name."
            )
        client.run(sys.argv[pytest_name_index + 1 :])


def _get_pattern_filters(config: Config) -> str | Callable[[str], bool]:
    """
    Jurigged takes in a pattern argument. The argument is either a glob string
    or a function that returns True if the path passed into it should be watched.

    This creates a function filter that will return True if the path matches.

    The logic takes in the --daemon-watch-globs and --daemon-ignore-watch-globs options
    and creates a function that will return True if the path matches the watch globs.

    The seen_paths set is used to prevent duplicate paths from being watched and also
    acts as a short circuit for paths that have already been seen.
    """
    global seen_paths

    import fnmatch
    import re

    def normalize(glob: str) -> str:
        if glob.startswith("~"):
            glob = os.path.expanduser(glob)
        elif not glob.startswith("/"):
            glob = os.path.abspath(glob)
        if os.path.isdir(glob):
            glob = os.path.join(glob, "*")
        return glob

    watch_globs = config.option.daemon_watch_globs.split(":")
    regex_matches = [re.compile(fnmatch.translate(normalize(glob))).match for glob in watch_globs]

    ignore_watch_globs = config.option.daemon_ignore_watch_globs
    if ignore_watch_globs:
        ignore_globs = config.option.daemon_ignore_watch_globs.split(":")
        ignore_regex_matches = [
            re.compile(fnmatch.translate(normalize(glob))).match for glob in ignore_globs
        ]
    else:
        ignore_regex_matches = []

    def matcher(filename) -> bool:
        if filename in seen_paths:
            return False
        seen_paths.add(filename)
        if any(regex_match(filename) for regex_match in regex_matches):
            if not any(
                ignore_regex_match(filename) for ignore_regex_match in ignore_regex_matches
            ):
                return True
        return False

    return matcher


def pytest_collection_modifyitems(session: Session, config: Config, items: list[Item]) -> None:
    """
    This hooks is called by pytest after the collection phase.

    This adds tests to the watch list automatically.

    The client should never get this far. This should only be
    used by the daemon.
    """
    global seen_paths
    import jurigged

    for item in items:
        if item.path and item.path not in seen_paths:
            jurigged.watch(pattern=str(item.path))
            seen_paths.add(item.path)



import copy
import os
import re
import socket
import subprocess
import sys
import time
from xmlrpc.server import SimpleXMLRPCServer

import pytest
from cachetools import TTLCache


class PytestDaemon:
    def __init__(self, daemon_host: str = "localhost", daemon_port: int = 4852) -> None:
        self._daemon_host = daemon_host
        self._daemon_port = daemon_port

    @property
    def pid_file(self) -> str:
        return f".pytest_hot_reloading_{self._daemon_port}.pid"

    @staticmethod
    def start(
        host: str,
        port: int,
        pytest_name: str = "pytest",
        watch_globs: str | None = None,
        ignore_watch_globs: str | None = None,
    ) -> None:
        # start the daemon such that it will not close when the parent process closes
        if host == "localhost":
            args = [
                sys.executable,
                "-m",
                pytest_name,
                "--daemon",
                "--daemon-port",
                str(port),
            ]
            if watch_globs:
                args += ["--daemon-watch-globs", watch_globs]
            if ignore_watch_globs:
                args += ["--daemon-ignore-watch-globs", ignore_watch_globs]
            subprocess.Popen(
                args,
                env=os.environ,
                cwd=os.getcwd(),
            )
        else:
            raise NotImplementedError("Only localhost is supported for now")
        PytestDaemon.wait_to_be_ready(host, port)

    @staticmethod
    def wait_to_be_ready(host: str = "localhost", port: int = 4852) -> None:
        # poll the connection to the daemon using sockets
        # and return when it is ready
        for _ in range(100):
            try:
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                sock.connect((host, port))
            except ConnectionRefusedError:
                time.sleep(0.1)
                continue
            else:
                break
        else:
            raise Exception("Could not connect to the daemon")

    def run_forever(self) -> None:  # create an XML-RPC server
        try:
            server = SimpleXMLRPCServer((self._daemon_host, self._daemon_port))
        except OSError as err:
            if "Address already in use" in str(err):
                self._kill_existing_daemon()
                time.sleep(2)
                server = SimpleXMLRPCServer((self._daemon_host, self._daemon_port))

        self._write_pid_file()

        # register the 'run_pytest' function
        server.register_function(self.run_pytest, "run_pytest")

        server.serve_forever()

    def _write_pid_file(self) -> None:
        with open(self.pid_file, "w") as f:
            f.write(str(os.getpid()))

    def _kill_existing_daemon(self) -> None:
        try:
            with open(self.pid_file, "r") as f:
                pid = int(f.read())
            os.kill(pid, 9)
        except FileNotFoundError:
            raise Exception(f"Port {self._daemon_port} is already in use")

    def run_pytest(self, args: list[str]) -> dict:
        # run pytest using command line args
        # run the pytest main logic

        self._workaround_library_issues(args)

        import pytest_hot_reloading.plugin as plugin

        # indicate to the plugin to NOT run custom pytest collect logic
        plugin.i_am_server = True

        # capture stdout and stderr
        # and return the output
        import io

        stdout = io.StringIO()
        stderr = io.StringIO()

        # backup originals
        stdout_bak = sys.stdout
        stderr_bak = sys.stderr

        sys.stdout = stdout
        sys.stderr = stderr

        import _pytest.main

        # monkeypatch in the main that does test collection caching
        orig_main = _pytest.main._main
        _pytest.main._main = _pytest_main

        try:
            # args must omit the calling program
            pytest.main(["--color=yes"] + args)
        finally:
            # restore originals
            _pytest.main._main = orig_main

            sys.stdout = stdout_bak
            sys.stderr = stderr_bak

            stdout.seek(0)
            stderr.seek(0)
            stdout_str = stdout.read()
            stderr_str = stderr.read()

            print(stdout_str, file=sys.stdout)
            print(stderr_str, file=sys.stderr)
        return {
            "stdout": self._remove_ansi_escape(stdout_str).encode("utf-8"),
            "stderr": self._remove_ansi_escape(stderr_str).encode("utf-8"),
        }

    def _remove_ansi_escape(self, s: str) -> str:
        return re.sub(r"\x1b(\[.*?[@-~]|\].*?(\x07|\x1b\\))", "", s, flags=re.MULTILINE)

    def _workaround_library_issues(self, args: list[str]) -> None:
        # load modules that workaround library issues, as needed
        pass


session_cache = TTLCache(16, 500)


def _pytest_main(config: pytest.Config, session: pytest.Session):
    """
    A monkey patched version of _pytest._main that caches test collection
    """
    import _pytest.capture

    _pytest.capture.CaptureManager.stop_global_capturing = lambda self: None
    start_global_capturing = _pytest.capture.CaptureManager.start_global_capturing
    resume_global_capture = _pytest.capture.CaptureManager.resume_global_capture

    def start_global_capture_if_needed(self: _pytest.capture.CaptureManager):
        if self._global_capturing is None:
            start_global_capturing(self)
        return resume_global_capture(self)

    _pytest.capture.CaptureManager.resume_global_capture = start_global_capture_if_needed

    def best_effort_copy(item, depth_remaining=2):
        """
        Copy test items. The items have references to modules and
        other things that cannot be deep copied.
        """
        if depth_remaining <= 0:
            return item
        try:
            item_copy = copy.copy(item)
        except TypeError:
            return item
        # NodeKeywords is an example of an object without a __dict__
        if hasattr(item, "__dict__"):
            for k, v in item.__dict__.items():
                try:
                    item_copy.__dict__[k] = copy.deepcopy(v)
                except KeyboardInterrupt:
                    raise
                except TypeError:
                    # Non-pickelable objects
                    item_copy.__dict__[k] = best_effort_copy(v, depth_remaining - 1)
        return item_copy

    # here config.args becomes basically the tests to run. Other arguments are omitted
    # not 100% sure this is always the case
    session_key = tuple(config.args)
    try:
        items = session_cache[session_key]
    except KeyError:
        # not in the cache, do test collection
        start = time.time()
        config.hook.pytest_collection(session=session)
        print(f"Pytest Daemon: Collection took {(time.time() - start):0.3f} seconds")
        session_cache[session_key] = tuple(best_effort_copy(x) for x in session.items)
    else:
        print("Pytest Daemon: Using cached collection")
        # Assign the prior test items (tests to run) and config to the current session
        session.items = items
        session.config = config
        for i in items:
            # Items have references to the config and the session
            i.config = config
            i.session = session
            if i._request:
                i._request._pyfuncitem = i
    config.hook.pytest_runtestloop(session=session)

    if session.testsfailed:
        return pytest.ExitCode.TESTS_FAILED
    elif session.testscollected == 0:
        return pytest.ExitCode.NO_TESTS_COLLECTED
    return None



import socket
import sys
import time
import xmlrpc.client


class PytestClient:
    _socket: socket.socket | None
    _daemon_host: str
    _daemon_port: int
    _pytest_name: str

    def __init__(
        self, daemon_host: str = "localhost", daemon_port: int = 4852, pytest_name: str = "pytest"
    ) -> None:
        self._socket = None
        self._daemon_host = daemon_host
        self._daemon_port = daemon_port
        self._pytest_name = pytest_name

    def run(self, args: list[str]) -> str:
        self._start_daemon_if_needed()

        server_url = f"http://{self._daemon_host}:{self._daemon_port}"
        server = xmlrpc.client.ServerProxy(server_url)

        start = time.time()
        result = server.run_pytest(args)
        print(f"Daemon took {(time.time() - start):.3f} seconds to reply")

        stdout = result["stdout"].data.decode("utf-8")
        stderr = result["stderr"].data.decode("utf-8")

        print(stdout, file=sys.stdout)
        print(stderr, file=sys.stderr)

    def abort(self) -> None:
        # Close the socket
        if self._socket:
            self._socket.close()

    def _start_daemon_if_needed(self) -> None:
        # check if the daemon is running on the expected host and port
        # if not, start the daemon

        # first, try to connect
        try:
            self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self._socket.connect((self._daemon_host, self._daemon_port))
            # the daemon is running
            # close the socket
            self._socket.close()
        except ConnectionRefusedError:
            # the daemon is not running
            # start the daemon
            self._start_daemon()

    def _start_daemon(self) -> None:
        from pytest_hot_reloading.daemon import PytestDaemon

        # start the daemon
        PytestDaemon.start(
            host=self._daemon_host, port=self._daemon_port, pytest_name=self._pytest_name
        )
